#!/usr/bin/env python
'''
component
Created by Seria at 25/11/2018 2:58 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from functools import partial
import tensorflow as tf
from copy import deepcopy
import os
from ..toolkit.utility import recordConfig

class Pod:
    def __init__(self, comp, symbol, name, msg=[]):
        self.component = comp
        self.symbol = symbol
        self.message = msg
        self.name = name

    def __rshift__(self, other):
        return Pod([self, other], '>', 'CASCADE')

    def __add__(self, other):
        return Pod([self, other], '+', 'ADD')

    def __sub__(self, other):
        return Pod([self, other], '-', 'SUB')

    def __mul__(self, other):
        return Pod([self, other], '*', 'MUL')

    def __matmul__(self, other):
        return Pod([self, other], '@', 'MATMUL')

    def __and__(self, other):
        return Pod([self, other], '&', 'CONCAT')

    def __pow__(self, power, modulo=None):
        assert isinstance(power, int)
        pod = deepcopy(self)
        pod._appendSerialNo('0')
        for p in range(1, power):
            pod_temp = deepcopy(self)
            pod_temp._appendSerialNo(str(p))
            pod = pod >> pod_temp
        return pod

    def _appendSerialNo(self, serial_no):
        if isinstance(self.component, list):
            self.component[0]._appendSerialNo(serial_no)
            self.component[1]._appendSerialNo(serial_no)
        else:
            self.name += '_' + serial_no

    def show(self):
        if isinstance(self.component, list):
            self.component[0].show()
            self.component[1].show()
        else:
            print(self.name)



class Component(object):
    def __init__(self):
        self.warehouse = ['CONV_1D', 'CONV', 'CONV_3D', 'CONV_TRANS', 'CONV_SEP', 'CONV_ATROUS',
                          'SIGMOID', 'TANH', 'SOFTMAX', 'RELU', 'RELU_LEAKY', 'RELU_EXP',
                          'MAX_POOL', 'AVG_POOL',
                          'DROPOUT', 'BATCH_NORM', 'EMBEDDING',
                          'RESHAPE', 'FLAT', 'SLICE',
                          'CLIP', 'DENSE',
                          'RESIZE',
                          'DUPLICATE', 'CONVERT',
                          'WEIGHT_DECAY', 'SIGM_XENTROPY', 'SFTM_XENTROPY', 'MSE', 'MAE',
                          'MOMENTUM', 'NESTEROV', 'ADAM']
        # Convolution
        self.CONV_1D = 0
        self.CONV = self.conv_2d
        self.CONV_3D = 2
        self.CONV_TRANS = 3
        self.CONV_SEP = 4
        self.CONV_ATROUS = 5
        # Activation
        self.SIGMOID = self.sigmoid
        self.TANH = self.tanh
        self.SOFTMAX = self.softmax
        self.RELU = self.relu
        self.RELU_LEAKY = self.relu_leaky
        self.RELU_EXP = 15
        # Pooling
        self.MAX_POOL = self.max_pool_2d
        self.AVG_POOL = self.avg_pool_2d
        # Distributing
        self.DROPOUT = self.dropout
        self.BATCH_NORM = self.batch_norm
        self.EMBEDDING = self.embedding
        # Reshape
        self.RESHAPE = self.reshape
        self.FLAT = self.flat
        self.SLICE = self.slice
        self.PAD = self.pad
        # Arithmetic
        self.CLIP = self.clip
        self.DENSE = self.dense
        self.ARGMAX = self.argmax
        # Image Manipulation
        self.RESIZE = self.resize
        # Copy or Rename
        self.DUPLICATE = self.duplicate
        self.CONVERT = self.convert
        # Loss
        self.WEIGHT_DECAY = self.weight_decay
        self.SIGM_XENTROPY = self.sigm_xentropy
        self.SFTM_XENTROPY = self.sftm_xentropy
        self.MSE = self.mse
        self.MAE = self.mae
        # Optimizer
        self.MOMENTUM = self.momentum
        self.NESTEROV = self.nesterov
        self.ADAM = self.adam
        # Metric
        self.ACC_CLS = self.accuracy_cls



    def addComp(self, name, comp, is_complete=False):
        name = name.upper()
        if name in self.warehouse:
            raise Exception('%s is an existing component in warehouse.' % name)
        if is_complete:
            def customizedComp(name):
                return Pod(partial(comp.component, name=name), comp.symbol, name, comp.message)
        else:
            assert not isinstance(comp.component, list)
            def customizedComp(**kwargs):
                symbol = kwargs.get(comp.symbol[1:], comp.symbol)
                return Pod(partial(comp.component, **kwargs), symbol, kwargs['name'], comp.message)
        exec('self.%s = customizedComp' % name)

    # ------------------------------------ Convolution ------------------------------------ #

    def _createVar(self, name, shape, initializer, param, regularizer=None):
        init_err = Exception('%s initializer is not defined or supported.' % initializer)
        # with tf.device('/gpu:0'):
        if initializer == 'xavier':
            if param is None:
                param = False
            var = tf.get_variable(name, shape, initializer=tf.contrib.layers.xavier_initializer(uniform=param))
        elif initializer == 'trunc_norm':
            if param is None:
                param = [0, 0.1]
            var = tf.get_variable(name, shape,
                                  initializer=tf.truncated_normal_initializer(mean=param[0], stddev=param[1]))
        elif initializer == 'rand_norm':
            if param is None:
                param = [0, 0.1]
            var = tf.get_variable(name,
                                  initializer=tf.random_normal(shape, mean=param[0], stddev=param[1]))
        elif initializer == 'zero':
            var = tf.get_variable(name, shape, initializer=tf.zeros_initializer())
        elif initializer == 'one':
            var = tf.get_variable(name, shape, initializer=tf.ones_initializer())
        else:
            raise init_err

        if regularizer == 'l2':
            weight_decay = tf.nn.l2_loss(var, name=name+'/regularizer')
            tf.add_to_collection(tf.GraphKeys.REGULARIZATION_LOSSES, weight_decay)
        elif regularizer is None:
            pass
        else:
            raise Exception('%s regularizer is not defined or supported.' % regularizer)

        return var

    def conv_2d(self, **kwargs):
        if 'input' in kwargs.keys():
            symbol = kwargs['input']
        else:
            symbol = '.input'
        return Pod(partial(self._conv_2d, **kwargs), symbol, kwargs['name'])
    def _conv_2d(self, name, input, kernel, out_chs, w_init='xavier', w_param=None,
                 b_init=None, b_param=None, w_reg='l2', b_reg='l2', stride=(1, 1), padding='same'):
        padding = padding.upper()
        in_chs = int(input.get_shape()[-1])
        in_out_chs = [in_chs, out_chs]
        w = self._createVar(name+'_w', kernel + in_out_chs, w_init, w_param, w_reg)
        if b_init:
            b = self._createVar(name+'_b', [in_out_chs[-1]], b_init, b_param, b_reg)
            return tf.nn.bias_add(tf.nn.conv2d(input, w, [1, stride[0], stride[1], 1], padding), b, name=name)
        else:
            return tf.nn.conv2d(input, w, [1, stride[0], stride[1], 1], padding, name=name)

    # ------------------------------------ Activation ------------------------------------ #

    def sigmoid(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._sigmoid, **kwargs), symbol, kwargs['name'])
    def _sigmoid(self, name, input):
        return tf.nn.sigmoid(input, name)

    def tanh(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._tanh, **kwargs), symbol, kwargs['name'])
    def _tanh(self, name, input):
        return tf.nn.tanh(input, name)

    def softmax(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._tanh, **kwargs), symbol, kwargs['name'])
    def _softmax(self, name, input):
        return tf.nn.softmax(input, name)

    def relu(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._relu, **kwargs), symbol, kwargs['name'])
    def _relu(self, name, input):
        return tf.nn.relu(input, name)

    def relu_leaky(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._relu_leaky, **kwargs), symbol, kwargs['name'])
    def _relu_leaky(self, name, input):
        return tf.nn.leaky_relu(input, name)

    # ------------------------------------ Distributing ------------------------------------ #

    def dropout(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._dropout, **kwargs), symbol, kwargs['name'])
    def _dropout(self, name, input, p_drop):
        return tf.identity(tf.nn.dropout(input, 1-p_drop), name=name)

    def batch_norm(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._batch_norm, **kwargs), symbol, kwargs['name'])
    def _batch_norm(self, name, input, is_train, beta=False, gamma=False):
        return tf.layers.batch_normalization(input, training=is_train, center=beta, scale=gamma, name=name)

    def embedding(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._embedding, **kwargs), symbol, kwargs['name'])
    def _embedding(self, name, input, vocabulary, vec_dims, w_init='xavier', w_param=None):
        embd_vec = self._createVar(name+'_vec', [vocabulary, vec_dims], w_init, w_param)
        return tf.nn.embedding_lookup(embd_vec, input, name=name)

    # ------------------------------------ Pooling ------------------------------------ #

    def max_pool_2d(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._max_pool_2d, **kwargs), symbol, kwargs['name'])
    def _max_pool_2d(self, name, input, kernel=(2, 2), stride=(2, 2), padding='same'):
        padding = padding.upper()
        return tf.nn.max_pool(input,
                              [1, kernel[0], kernel[1], 1],
                              [1, stride[0], stride[1], 1],
                              padding, name=name)

    def avg_pool_2d(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._avg_pool_2d, **kwargs), symbol, kwargs['name'])
    def _avg_pool_2d(self, name, input, kernel=(2, 2), stride=(2, 2), padding='same'):
        padding = padding.upper()
        return tf.nn.avg_pool(input,
                              [1, kernel[0], kernel[1], 1],
                              [1, stride[0], stride[1], 1],
                              padding, name=name)

    # ------------------------------------ Reshape ------------------------------------ #

    def reshape(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._reshape, **kwargs), symbol, kwargs['name'])
    def _reshape(self, name, input, shape):
        return tf.reshape(input, shape, name)

    def flat(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._flat, **kwargs), symbol, kwargs['name'])
    def _flat(self, name, input):
        return tf.layers.flatten(input, name=name)

    def slice(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._slice, **kwargs), symbol, kwargs['name'])
    def _slice(self, name, input, indices):
        for d in range(len(indices)):
            output = input[indices[d][0]:indices[d][1]]
        return tf.identity(output, name)

    def pad(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._pad, **kwargs), symbol, kwargs['name'])
    def _pad(self, name, input, margin, fill_in=0):
        return tf.pad(input, margin, constant_values=fill_in, name=name)

    # -------------------------------------- Arithmetic -------------------------------------- #

    def clip(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._clip, **kwargs), symbol, kwargs['name'])
    def _clip(self, name, input, min_max_vals):
        return tf.clip_by_value(input, min_max_vals[0], min_max_vals[1], name)

    def dense(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._dense, **kwargs), symbol, kwargs['name'])
    def _dense(self, name, input, out_chs, w_init='xavier', w_param=None,
               b_init='zero', b_param=None, w_reg='l2', b_reg='l2'):
        in_chs = int(input.get_shape()[-1])
        in_out_chs = [in_chs, out_chs]
        w = self._createVar(name + '_w', in_out_chs, w_init, w_param, w_reg)
        if b_init:
            b = self._createVar(name + '_b', [in_out_chs[-1]], b_init, b_param, b_reg)
            return tf.nn.bias_add(tf.matmul(input, w), b, name=name)
        else:
            return tf.matmul(input, w, name=name)

    def argmax(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._argmax, **kwargs), symbol, kwargs['name'])
    def _argmax(self, name, input, axis):
        return tf.argmax(input, axis, name=name)

    # ------------------------------------ Image Manipulation ------------------------------------ #

    def resize(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._resize, **kwargs), symbol, kwargs['name'])
    def _resize(self, name, input, size, method='bilinear'):
        if method == 'bilinear':
            return tf.image.resize_bilinear(input, size, name=name)
        elif method == 'bicubic':
            return tf.image.resize_bicubic(input, size, name=name)
        elif method == 'crop':
            return tf.image.resize_image_with_crop_or_pad(input, size, name=name)
        else:
            raise KeyError('%s is not a legal resize method.' % method)


    # ------------------------------------ Redefine or Rename ------------------------------------ #

    def duplicate(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._duplicate, **kwargs), symbol, kwargs['name'])
    def _duplicate(self, name, input):
        return tf.identity(input, name)

    def convert(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._convert, **kwargs), symbol, kwargs['name'])
    def _convert(self, name, input, dtype, trainable=False):
        '''
        convert data type or convert list/numpy array to tensor
        :param name:
        :param input: input tensor / list / numpy array
        :param dtype:
        :param trainable: if tensor is trainable
        :return: tensor
        '''
        if isinstance(input, (tf.Tensor, tf.SparseTensor, tf.Variable)):
            return tf.cast(input, tf.as_dtype(dtype), name=name)
        else:
            return tf.Variable(input, trainable=trainable, name=name)

    # ------------------------------------ Loss ------------------------------------ #

    def weight_decay(self, **kwargs):
        if 'penalty' in kwargs.keys():
            message = [kwargs['penalty']]
        else:
            message = []
        return Pod(partial(self._weight_decay, **kwargs), '.penalty', kwargs['name'], message)
    def _weight_decay(self, name, penalty, decay_scope=None):
        if decay_scope is None:
            return tf.multiply(tf.add_n(tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES)),
                               penalty, name=name)
        else:
            return tf.multiply(tf.add_n(tf.get_collection(tf.GraphKeys.REGULARIZATION_LOSSES, scope=decay_scope)),
                               penalty, name=name)

    def sigm_xentropy(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._sigm_xentropy, **kwargs), symbol, kwargs['name'], [kwargs['label']])
    def _sigm_xentropy(self, name, input, label):
        labels = tf.cast(label, tf.float32)
        return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=label, logits=input, name=name))

    def sftm_xentropy(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._sftm_xentropy, **kwargs), symbol, kwargs['name'], [kwargs['label']])
    def _sftm_xentropy(self, name, input, label, is_one_hot):
        if is_one_hot:
            labels = tf.cast(label, tf.float32)
            return tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(labels=labels, logits=input, name=name))
        else:
            return tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(
                                        labels=label, logits= input, name=name))

    def mse(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._mse, **kwargs), symbol, kwargs['name'], [kwargs['label']])
    def _mse(self, name, input, label):
        return tf.reduce_mean(tf.squared_difference(input, label), name=name)

    def mae(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._mae, **kwargs), symbol, kwargs['name'], [kwargs['label']])
    def _mae(self, name, input, label):
        return tf.reduce_mean(tf.abs(input-label), name=name)

    # ------------------------------------ Optimizer ------------------------------------ #

    def _lrStrategy(self, mile_meter, lr, lr_decay, miles, param):
        if lr_decay == 'exp':
            return tf.train.exponential_decay(lr, mile_meter, miles, param, staircase=False)
        elif lr_decay == 'exp_stair':
            return tf.train.exponential_decay(lr, mile_meter, miles, param, staircase=True)
        elif lr_decay == 'poly':
            return tf.train.polynomial_decay(lr, mile_meter, miles, power=param, cycle=False)
        elif lr_decay == 'poly_cycle':
            return tf.train.polynomial_decay(lr, mile_meter, miles, power=param, cycle=True)
        elif lr_decay == 'stair':
            return tf.train.piecewise_constant(mile_meter, [i*param for i in range(1, len(lr))], lr)
        else:
            raise KeyError('%s decay is not supported or defined.' % lr_decay)

    def _computeGrad(self, optz, cost, update_scope, ignore_name, grad_limit):
        update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
        with tf.control_dependencies(update_ops):
            if update_scope is None:
                tr_var = tf.trainable_variables()
            else:
                tr_var = tf.trainable_variables(scope=update_scope)
            if not ignore_name is None:
                update_var = []
                for tv in tr_var:
                    flag_ignore = False
                    for ig in ignore_name:
                        if ig in tv.op.name:
                            flag_ignore = True
                            break
                    if not flag_ignore:
                        update_var.append(tv)
            else:
                update_var = tr_var
            grad_var_pairs = optz.compute_gradients(cost, var_list=update_var)

            if grad_limit is None:
                return grad_var_pairs
            else:
                return [(tf.clip_by_value(grad, -grad_limit, grad_limit), var) for grad, var in grad_var_pairs]

    def momentum(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._momentum, **kwargs), symbol, kwargs['name'])
    def _momentum(self, name, input, mile_meter, lr, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None):
        recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
                     {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
                      'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit})
        if isinstance(lr_decay, str):
            lr = self._lrStrategy(mile_meter, lr, lr_decay, lr_miles, decay_param)
        optz = tf.train.MomentumOptimizer(lr, mmnt)
        grad = self._computeGrad(optz, input, update_scope, ignore_name, grad_limit)
        train_op = optz.apply_gradients(grad, mile_meter, name=name)
        return train_op

    def nesterov(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._nesterov, **kwargs), symbol, kwargs['name'])
    def _nesterov(self, name, input, mile_meter, lr, mmnt=0.9, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None):
        recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
                     {'lr': lr, 'mmnt': mmnt, 'update_scope': update_scope, 'ignore_name': ignore_name,
                      'lr_decay': lr_decay, 'lr_miles': lr_miles, 'decay_param': decay_param, 'grad_limit': grad_limit})
        if isinstance(lr_decay, str):
            lr = self._lrStrategy(mile_meter, lr, lr_decay, lr_miles, decay_param)
        optz = tf.train.MomentumOptimizer(lr, mmnt, use_nesterov=True)
        grad = self._computeGrad(optz, input, update_scope, ignore_name, grad_limit)
        train_op = optz.apply_gradients(grad, mile_meter, name=name)
        return train_op

    def adam(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._adam, **kwargs), symbol, kwargs['name'])
    def _adam(self, name, input, mile_meter, lr, mmnt1=0.9, mmnt2=0.999, update_scope=None, ignore_name=None,
                  lr_decay=None, lr_miles=None, decay_param=None, grad_limit=None):
        recordConfig(os.path.join(os.getcwd(), 'temp_config.json'),
                     {'lr': lr, 'mmnt1': mmnt1, 'mmnt2': mmnt2, 'update_scope': update_scope,
                      'ignore_name': ignore_name, 'lr_decay': lr_decay, 'lr_miles': lr_miles,
                      'decay_param': decay_param, 'grad_limit': grad_limit})
        if isinstance(lr_decay, str):
            lr = self._lrStrategy(mile_meter, lr, lr_decay, lr_miles, decay_param)
        optz = tf.train.AdamOptimizer(lr, beta1=mmnt1, beta2=mmnt2)
        grad = self._computeGrad(optz, input, update_scope, ignore_name, grad_limit)
        train_op = optz.apply_gradients(grad, mile_meter, name=name)
        return train_op

    # ------------------------------------ Metric ------------------------------------ #

    def accuracy_cls(self, **kwargs):
        symbol = kwargs.get('input', '.input')
        return Pod(partial(self._accuracy_cls, **kwargs), symbol, kwargs['name'], [kwargs['label']])
    def _accuracy_cls(self, name, input, label, multi_class, is_one_hot=False):
        if multi_class:
            assert not is_one_hot
            pred = tf.cast(tf.round(input), label.dtype)
            correct = tf.reduce_all(tf.equal(pred, label), axis=-1)
            return tf.reduce_mean(tf.cast(correct, tf.float32), name=name)
        else:
            if is_one_hot:
                label = tf.argmax(label, axis=-1)
            pred = tf.argmax(input, axis=-1)
            correct = tf.equal(pred, label)
            return tf.reduce_mean(tf.cast(correct, tf.float32), name=name)


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import math
import io
import random as rand
import numpy as np
import h5py
from PIL import Image
import skimage.transform as imgtrans

from ..toolkit.decorator import Timer

class FuelTank(object):
    def __init__(self, param):
        self.param = param
        self.epoch = 0
        duration = self._loadData()
        print('+' + (61 * '-') + '+')
        print('| \033[1;35m%-17s\033[0m fuel tank has been mounted within \033[1;35m%6.3fs\033[0m |'
              % (param['name'], duration))
        print('+' + (61 * '-') + '+')

    def __del__(self):
        self.datafile.close()
        print('+' + (50 * '-') + '+')
        print('| \033[1;35m%-17s\033[0m fuel tank is no longer mounted |' % self.param['name'])
        print('+' + (50 * '-') + '+')

    def _shuffleData(self):
        '''
        Args:
        num: batch size
        leftover: the rest samples of last epoch
        '''
        nsample = self.nsample
        for j in range(nsample):
            idx = int(rand.random() * (nsample - j)) + j
            temp = self.order[j]
            self.order[j] = self.order[idx]
            self.order[idx] = temp

    def _dataAugmentation(self, source):
        def _spatialAugmentation(img, da_type, theta):
            data_aug_err = Exception('%s spatial augmentation is unsupported.' % da_type)
            if da_type == 'flip':
                aug_img = np.zeros(img.shape)
                if self.param['channel'] == 1:
                    aug_img = np.fliplr(img)
                else:
                    for c in range(self.param['channel']):
                        aug_img[:, :, c] = np.fliplr(img[:, :, c])
                return aug_img
            elif da_type == 'crop':
                # theta: ((area_ratio_inf, area_ratio_sup), (aspect_ratio_inf, aspect_ratio_sup))
                sqrt_aspect_ratio = math.sqrt(rand.random() * (theta[1][1]-theta[1][0]) + theta[1][0])
                sqrt_area_ratio = math.sqrt(rand.random() * (theta[0][1]-theta[0][0]) + theta[0][0])
                w = int(sqrt_area_ratio*self.param['width']*sqrt_aspect_ratio)
                h = int(sqrt_area_ratio*self.param['height']/sqrt_aspect_ratio)
                x = int(rand.random() * (self.param['width']-w+1))
                y = int(rand.random() * (self.param['height']-h+1))
                # import pdb
                # pdb.set_trace()
                aug_img = imgtrans.resize(img[y:y+h, x:x+w].copy(), (self.param['height'], self.param['width']))
                return aug_img
            elif da_type == 'brightness':
                perturbation = rand.random() * (theta[1]-theta[0]) +theta[0]
                aug_img = perturbation + img  # pixels range between [theta, 1+theta]
                aug_img = np.clip(aug_img, 0, 1)  # pixels range between [0, 1]
                return aug_img
            elif da_type == 'gamma_contrast':
                perturbation = rand.random() * (theta[1] - theta[0]) + theta[0]
                aug_img = perturbation * (img - 0.5) # pixels range between [0, theta]
                aug_img = np.clip(aug_img, -0.5, 0.5) + 0.5 # pixels range between [0, 1]
                return aug_img
            elif da_type == 'log_contrast':
                perturbation = rand.random() * (theta[1] - theta[0]) + theta[0]
                aug_img = perturbation * np.log(img + 1) # pixels range between [0, theta * ln2]
                aug_img = np.where(aug_img > 1, 1, aug_img)  # pixels range between [0, 1]
                return aug_img
            else:
                raise data_aug_err

        def _temporalAugmentation(seq, da_type, theta):
            data_aug_err = Exception('%s temporal augmentation is unsupported.' % da_type)
            if da_type == 'sample':
                seq_len = self.param['frame'] // theta # here theta is sampling frequency
                aug_seq = np.zeros((seq_len,) + seq.shape[1:])
                for i in range(seq_len):
                    aug_seq[i] = seq[int((i + rand.random()) * theta)]
                return aug_seq
            else:
                raise data_aug_err

        if self.param['frame'] != 0: # if this call is for sequential data augmentation
            for t,ta in enumerate(self.param['temporal_aug'].split(',')):
                if not ta:
                    continue
                if rand.random() < self.param['p_ta'][t]:
                    source = _temporalAugmentation(source, ta, self.param['theta_ta'][t])
                    for s, sa in enumerate(self.param['spatial_aug'].split(',')):
                        if not sa:
                            continue
                        if rand.random() < self.param['p_sa'][s]:
                            for i in range(source.shape[0]):
                                source[i] = _spatialAugmentation(source[i], sa, self.param['theta_sa'][t])
        else:
            for s,sa in enumerate(self.param['spatial_aug'].split(',')):
                if not sa:
                    continue
                if rand.random() < self.param['p_sa'][s]:
                    source = _spatialAugmentation(source, sa, self.param['theta_sa'][s])

        return source * 2 - 1 # pixels range between [-1, 1]

    @Timer
    def _loadData(self):
        # --------------------------------- load data --------------------------------- #
        format = self.param['data_path'].split('.')[-1]
        if format == 'npz':
            self.datafile = np.load(self.param['data_path'])

        elif format == 'hdf5':
            self.datafile = h5py.File(self.param['data_path'], 'r')


        self.attributes = self.datafile.keys()
        self.dtypes = {}
        if self.param['frame'] == -1:
            self.param['frame'] = self.datafile['max_frames']
            self.attributes.remove('max_frames')
        for atr in self.attributes:
            if atr == self.param['data_key'] and self.param['is_compressed']:
                self.dtypes[atr] = np.dtype('float32')
            else:
                self.dtypes[atr] = self.datafile[atr].dtype

        # get number of samples and channels
        self.nsample = self.datafile[self.param['data_key']].shape[0]

        # --------------------------- initialize and shuffle --------------------------- #
        order = [i for i in range(self.nsample)]
        self.curr_idx = 0
        self.order = order
        if self.param['if_shuffle']:
            self._shuffleData()

    def _byte2Array(self, data_src):
        if not self.param['is_compressed']: # no need to recover data from bytes
            return data_src
        data_bytes = data_src.tobytes()
        data_pil = Image.open(io.BytesIO(data_bytes))
        if self.param['width']>0 and self.param['height']>0:
            # if need lower the image resolution
            if self.param['resol_ratio'] < 1:
                data_pil = data_pil.resize((int(self.param['resol_ratio'] * self.param['width']),
                                            int(self.param['resol_ratio'] * self.param['height'])))
            data_pil = data_pil.resize((self.param['width'], self.param['height']))
        data_np = np.array(data_pil) # pixels range between [0, 255]
        data_np = (data_np/255).astype(np.float32) # pixels range between [0, 1]
        if self.param['channel'] == 1:
            data_np = np.expand_dims(data_np, -1)
        return data_np

    def _fetchBatches(self):
        '''
        Fetch next batch
        '''
        while 1:
            self.curr_idx += self.param['batch_size']
            curr_batch = {}
            for atr in self.attributes:
                curr_batch[atr] = []
            # start a new epoch
            low_power = False
            # import time
            # start = time.time()
            if self.curr_idx > self.nsample:
                low_power = True
                self.epoch = -self.epoch
                # if the samples left are not enough to feed a mini batch, let's say n samples.
                # the current batch should be the last n samples in addition to the first (bs-n) samples
                for idx in self.order[self.curr_idx - self.param['batch_size']:self.nsample]:
                    for atr in self.attributes:
                        if atr == self.param['data_key']:
                            curr_sample = self.datafile[atr][idx]
                            if self.param['frame'] > 0: # if this is sequential data
                                temp_data = np.zeros((self.param['frame'], self.param['height'],
                                                      self.param['width'], self.param['channel']))
                                for f in range(curr_sample.shape[0]):
                                    if f < self.param['frame']:
                                        temp_data[f] = self._byte2Array(curr_sample[f])
                                    else:
                                        break
                                curr_batch[atr] += [self._dataAugmentation(temp_data)]
                            else:
                                curr_batch[atr] += [self._dataAugmentation(self._byte2Array(curr_sample))]
                        else:
                            curr_batch[atr] += [self.datafile[atr][idx]]
                for idx in self.order[0:self.curr_idx - self.nsample]:
                    for atr in self.attributes:
                        if atr == self.param['data_key']:
                            curr_batch[atr] += [self._dataAugmentation(self._byte2Array(self.datafile[atr][idx]))]
                        else:
                            curr_batch[atr] += [self.datafile[atr][idx]]

                if self.param['if_shuffle']:
                    # reshuffle data to read
                    self._shuffleData()
                else:
                    self.order = [i for i in range(self.nsample)]
                self.curr_idx = 0
            else:
                for idx in self.order[self.curr_idx - self.param['batch_size']:self.curr_idx]:
                    for atr in self.attributes:
                        if atr == self.param['data_key']:
                            curr_batch[atr] += [self._dataAugmentation(self._byte2Array(self.datafile[atr][idx]))]
                        else:
                            curr_batch[atr] += [self.datafile[atr][idx]]

            for atr in self.attributes:
                curr_batch[atr] = np.array(curr_batch[atr]).astype(self.dtypes[atr])
            # print('%.3fs'%(time.time()-start))
            yield curr_batch
            if low_power:
                # epoch counter turns positive when an epoch ends
                self.epoch = abs(self.epoch) + 1


#!/usr/bin/env python
'''
utility
Created by Seria at 14/11/2018 8:33 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import json
import numpy as np
import subprocess as subp

def toDenseLabel(labels, nclass, on_value=1, off_value=0):
    batch_size = len(labels)
    # initialize dense labels
    one_hot = off_value * np.ones((batch_size * nclass))
    indices = []
    if isinstance(labels[0], str):
        for b in range(batch_size):
            indices += [int(s) + b * nclass for s in labels[b].split(',')]
    else: # labels is a nested list
        for b in range(batch_size):
            indices += [l + b * nclass for l in labels[b]]
    one_hot[indices] = on_value
    return np.reshape(one_hot, (batch_size, nclass))


def getAvailabelGPU(least_mem):
    p = subp.Popen('nvidia-smi', stdout=subp.PIPE)
    gpu_id = 0  # next gpu we are about to probe
    flag_gpu = False
    max_occupied = least_mem
    id_best = -1  # gpu having max avialable memory
    for l in p.stdout.readlines():
        line = l.decode('utf-8').split()
        if len(line) < 1:
            break
        elif len(line) < 2:
            continue
        if line[1] == str(gpu_id):
            flag_gpu = True
            continue
        if flag_gpu:
            occupancy = int(line[8].split('M')[0])
            if occupancy < max_occupied:
                max_occupied = occupancy
                id_best = gpu_id
            gpu_id += 1
            flag_gpu = False
    return id_best

def parseConfig(config_path):
    with open(config_path, 'r') as config_file:
        return json.load(config_file)

def recordConfig(config_path, config):
    with open(config_path, 'w') as config_file:
        json.dump(config, config_file, indent=4)


#!/usr/bin/env python
'''
ugly
Created by Seria at 14/12/2018 8:06 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========

'''
# -*- coding:utf-8 -*-
import nebulae
import tensorflow as tf
import os
import csv
import h5py
import numpy as np
import pandas as pd
import scipy.misc as misc
import matplotlib.pyplot as plt
import resv2


fd = nebulae.fuel.FuelDepot()
fname = 'kpi'
fd.loadFuel(name=fname + '-ct',
            batch_size=32,
            height=224,
            width=224,
            channel=3,
            data_path='/home/ai/seria/data/kpi18_mxt.hdf5',
            data_key='image',
            spatial_aug='flip,crop',
            p_sa=(0.5, 0.5),
            theta_sa=(0, ((0.25, 1), (3/4, 4/3))),
            if_shuffle=False)
fd.loadFuel(name=fname + '-cv',
            batch_size=32,
            height=224,
            width=224,
            channel=3,
            data_path='/home/ai/seria/data/kpi18_clv.hdf5',
            data_key='image',
            if_shuffle=False)
fd.loadFuel(name=fname + '-test',
            batch_size=32,
            height=224,
            width=224,
            channel=3,
            data_path='/home/ai/seria/data/kpi18_test.hdf5',
            data_key='image',
            if_shuffle=False)
# fd.loadFuel(name=fname + '-n',
#             batch_size=50,
#             height=224,
#             width=224,
#             channel=3,
#             data_path='/home/ai/seria/data/kpi18_trn.hdf5',
#             data_key='image',
#             if_shuffle=False)

# launcher = tf.placeholder(tf.float32, (None, 224, 224, 3), 'FL/input')
# terminator = tf.placeholder(tf.int64, (None), 'FL/label')

COMP = nebulae.spacedock.Component()
ls = nebulae.aerolog.LayoutSheet('/home/ai/seria/kpi_ug')
scope = 'sc'
mdtype = 'resnet_v2_152' ######################################################################
sc = nebulae.spacedock.SpaceCraft(scope, layout_sheet=ls)
sc.fuelLine('input', (None, 224, 224, 3), 'float32')
sc.fuelLine('label', (None), 'int64')
sc.fuelLine('is_train', (), 'bool', default=False)
with tf.contrib.slim.arg_scope(resv2.resnet_arg_scope()):
    logits, endpts = resv2.resnet_v2_152(sc.layout['input'], is_training=sc.layout['is_train'],
                                    num_classes=128) ###########################################################
feature = tf.squeeze(endpts['global_pool'])
# cost = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(
#             logits=logits, labels=sc.layout['label']))
#
# prob = sc.assemble(COMP.SOFTMAX(name='sftm1', input=logits))
# mmnt = COMP.ADAM(name='adam', input=cost, mile_meter=sc.miles,
#                      lr=1e-6, update_scope=mdtype, lr_decay='exp_stair',
#                      lr_miles=1000, decay_param=0.95, grad_limit=4.)
# optz = sc.assemble(mmnt)
#
# correct = tf.equal(tf.argmax(sc.layout[prob], 1), sc.layout['label'])
# accuracy = tf.reduce_mean(tf.cast(correct, tf.float32))

ind = tf.argmax(tf.nn.softmax(logits), 1)
# error = tf.where(tf.logical_not(correct))
# error = tf.where(correct)


config_proto = tf.ConfigProto(allow_soft_placement=True)
config_proto.gpu_options.per_process_gpu_memory_fraction = 0.99
config_proto.gpu_options.visible_device_list = '0' ###################################################
config_proto.gpu_options.allow_growth = True
sess = tf.Session(config=config_proto)
sess.run(tf.global_variables_initializer())

mdir = '/home/ai/seria/model/res-e' ############################################################
if not os.path.exists(mdir):
    os.mkdir(mdir)


ckpt = tf.train.latest_checkpoint(mdir)
vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES)
# vars = tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=mdtype)
# vars = [v for v in vars if 'logits' not in v.name]
restorer = tf.train.Saver(vars)
restorer.restore(sess, ckpt)
# restorer.restore(sess, '/home/ai/seria/model/res-slim/%s.ckpt'%mdtype)


# saver = tf.train.Saver(tf.global_variables(), max_to_keep=1)
#
# epoch = 7
# mile = fd.milesPerEpoch(fname + '-ct')
# miles = epoch*mile+1
# monitor = np.zeros((100,))
# record_t = miles*[0]
# record_v = epoch*[0]
# for s in range(1):
#     batch = fd.nextBatch(fname + '-ct')
#     _, acc, loss = sess.run([sc.layout[optz], accuracy, cost],
#                             feed_dict={sc.layout['input']: batch['image'],
#                                        sc.layout['label']: batch['label'],
#                                        sc.layout['is_train']: True})
#     monitor[s % 100] = acc
#     avg = monitor.mean()
#     record_t[s] = avg*100
#     # import pdb
#     # pdb.set_trace()
#     if s % 50 == 0:
#         print('epoch #%d - %d miles acc: %.2f%%, loss: %.4f' %
#               (fd.currentEpoch(fname + '-ct'), s, acc * 100, loss))
#         if s % 200 == 0:
#             saver.save(sess, os.path.join(mdir,'res_n'), global_step=s, write_meta_graph=True) ##########################
#     if s>0 and s%fd.milesPerEpoch(fname + '-ct') == 0:
#         print(50 * '=')
#         acc_tot = 0
#         loss_tot = 0
#         steps = fd.milesPerEpoch(fname + '-cv')
#         for m in range(steps):
#             batch = fd.nextBatch(fname + '-cv')
#             acc, loss = sess.run([accuracy, cost], feed_dict={sc.layout['input']: 255*batch['image']-np.array([123.68, 116.779, 103.939]),
#                                                               sc.layout['label']: batch['label'],
#                                                               })
#             # import pdb
#             # pdb.set_trace()
#             # misc.imsave('/home/ai/seria/err.jpg', batch['image'][0])
#             acc_tot += acc
#             loss_tot += loss
#         print('epoch #%d: acc: %.2f%%, loss: %.4f' % (
#         fd.currentEpoch(fname + '-cv'), acc_tot / steps * 100, loss_tot / steps))
#         print(50 * '=')
#         record_v[fd.currentEpoch(fname + '-cv')-1] = acc_tot / steps * 100
# plt.switch_backend('agg')
# plt.plot(np.arange(miles), record_t, c='blue')
# plt.plot(np.arange(mile,miles,mile), record_v, c='red')
# plt.savefig(os.path.join(mdir,'curve.jpg'))
#
#
# fd.unloadFuel(fname + '-ct')
# fd.unloadFuel(fname + '-cv')


steps = fd.milesPerEpoch(fname + '-ct')
vec = np.zeros((40853,2048))
for m in range(steps):
    if m==steps-1:
        fd.editProperty(fname + '-ct', batch_size=21)
    batch = fd.nextBatch(fname + '-ct')
    feat = sess.run(feature, feed_dict={sc.layout['input']: batch['image'],
                                        sc.layout['is_train']: True})
    if m == steps - 1:
        vec[-21:] = feat
    else:
        vec[m * 32:(m + 1) * 32] = feat
hdf5 = h5py.File('/home/ai/seria/mxt.hdf5', 'w')
hdf5['train'] = vec
hdf5.close()

steps = fd.milesPerEpoch(fname + '-cv')
vec = np.zeros((4096,2048))
for m in range(steps):
    batch = fd.nextBatch(fname + '-cv')
    feat = sess.run(feature, feed_dict={sc.layout['input']: batch['image'],
                                        sc.layout['is_train']: True})
    vec[m * 32:(m + 1) * 32] = feat
hdf5 = h5py.File('/home/ai/seria/clv.hdf5', 'w')
hdf5['val'] = vec
hdf5.close()


steps = fd.milesPerEpoch(fname + '-test')
vec = np.zeros((19200,2048))
indices = np.zeros((19200,))
for m in range(steps):
    batch = fd.nextBatch(fname + '-test')
    idx,feat = sess.run([ind,feature], feed_dict={sc.layout['input']: batch['image'],
                                                  sc.layout['is_train']: True})
    indices[m * 32:(m + 1) * 32] = idx
    vec[m * 32:(m + 1) * 32] = feat
flist = os.listdir('/home/ai/challenge/test')
flist.remove('kpi18_test.csv')
with open('/home/ai/seria/liyuyulv.csv', 'w') as csvf:
    csvw = csv.writer(csvf)
    for i, f in enumerate(flist):
        csvw.writerow([f, str(indices[i] + 1)])
hdf5 = h5py.File('/home/ai/seria/test.hdf5', 'w')
hdf5['test'] = vec
hdf5.close()

fd.unloadFuel(fname + '-ct')
fd.unloadFuel(fname + '-cv')
fd.unloadFuel(fname + '-test')


# steps = fd.milesPerEpoch(fname + '-test')
# indices = np.zeros((19200,))
# for m in range(steps):
#     batch = fd.nextBatch(fname + '-test')
#     idx = sess.run(ind, feed_dict={sc.layout['input']: batch['image'],
#                                                       })
#     indices[m*32:(m+1)*32] = idx
# flist = os.listdir('/home/ai/challenge/test')
# flist.remove('kpi18_test.csv')
# with open('/home/ai/seria/liyuyulv.csv', 'w') as csvf:
#     csvw = csv.writer(csvf)
#     for i,f in enumerate(flist):
#         csvw.writerow([f,str(indices[i]+1)])
#
# fd.unloadFuel(fname + '-test')



# steps = fd.milesPerEpoch(fname + '-n')
# indices = []
# for m in range(steps):
#     batch = fd.nextBatch(fname + '-n')
#     err = sess.run(error, feed_dict={sc.layout['input']: batch['image'],
#                                    sc.layout['label']: batch['label'],
#                                      sc.layout['is_train']: True})
#     for e in err.flatten():
#         indices.append(e+m*50)
# flist = pd.read_csv('/home/ai/seria/kpi18_trn.csv', header=0)
# with open('/home/ai/seria/model/reliable.csv', 'w') as csvf:
#     csvw = csv.writer(csvf)
#     for i in indices:
#         csvw.writerow([flist.iloc[i,0], flist.iloc[i,1]])
#
# fd.unloadFuel(fname + '-n')